#[doc = include_str!("../README.md")]
use jsonschema::{Draft, JSONSchema};
use serde_json::{Map, Value};

pub mod error;

pub use error::*;

/// The JSON schema inspector (validator)
///
/// This contains the necessary glue to make all the schemas from the OpenAPI spec (tested on 3.0.3) work with the `jsonschema` crate.
/// This is achieved by simply resolving the local references present in the schemas with the actual schemas.
///
/// It is recommended to initialize one `SchemaInspector` for one schema only once, and use the same instance to validate/inspect
/// multiple JSON objects to avoid performance overhead of compiling the schema multiple times.
///
/// Note: Resolving references is optional and is not required for schemas that do not contain any references to nested schemas.
#[derive(Debug)]
pub struct SchemaInspector {
    schema: JSONSchema, // the compiled schema used to validate JSON
}

impl SchemaInspector {
    const REF: &'static str = "$ref";

    /// Initialize the `SchemaInspector` using the name of the type provided to the function using the turbofish syntax instead of using a string
    ///
    /// Note: Resolving references is optional and is not required for schemas that do not contain any references to nested schemas.
    /// It is recommended to set `resolve_refs` to false when there are no references present in the schema to avoid extra overhead.
    ///
    /// ## Example
    /// ```
    /// use schema_police::{SchemaInspector, InitError};
    ///
    /// fn construct_inspector(schema: &serde_json::Value) -> Result<SchemaInspector, InitError> {
    ///     SchemaInspector::new_infer::<SchemaType>(schema, true)
    /// }
    ///
    /// #[derive(utoipa::ToSchema)]
    /// struct SchemaType {
    ///     field: String,
    /// }
    /// ```
    ///
    /// Note: the function internally uses [`std::any::type_name`] to get the name of the type. Accodring to the std lib docs, the function
    /// is meant to be used for debugging and makes no guarantees. Therefore, it is not recommended to rely on this method in mission critical contexts.
    pub fn new_infer<T>(schemas: &Value, resolve_refs: bool) -> Result<Self, InitError> {
        // get target name from type inference
        let target = std::any::type_name::<T>()
            .split("::")
            .last()
            .ok_or(InitError::InferenceFailure)?;
        Self::new(target, schemas, resolve_refs)
    }

    /// Construct an instance of the `SchemaInspector` using the provided OpenAPI spec's schemas (in the form of `serde_json::Value`)
    ///
    /// Note: Resolving references is optional and is not required for schemas that do not contain any references to nested schemas.
    /// It is recommended to set `resolve_refs` to false when there are no references present in the schema to avoid extra overhead.
    pub fn new(target: &str, schemas: &Value, resolve_refs: bool) -> Result<Self, InitError> {
        use InitError::*;
        // extract schemas object
        let Value::Object(schemas) = schemas else {
            return Err(InitError::InvalidSchema(None));
        };
        // extract target from schema
        let mut target = schemas
            .get(target)
            .ok_or_else(|| SchemaNotFound(target.to_owned()))?
            .to_owned();
        if resolve_refs {
            Self::resolve_references(schemas, &mut target)?;
        }
        Ok(Self {
            schema: JSONSchema::options()
                .with_draft(Draft::Draft202012)
                .compile(&target)
                .map_err(|_| CompileFailure)?,
        })
    }

    /// Replace internal '$ref' references with the correct schemas
    /// The current implementation disregards the paths in the references and only uses the schema names
    /// to retrieve them from the root schemas object
    fn resolve_references(schemas: &Map<String, Value>, val: &mut Value) -> Result<(), InitError> {
        if let Some(target) = Self::extract_ref(val) {
            let mut resolved_schema = schemas
                .get(&target)
                .ok_or(InitError::ResolutionFailure(format!(
                    "{target} not found in the given schemas"
                )))?
                .clone();
            // resolve any references within the resolved schema
            Self::resolve_references(schemas, &mut resolved_schema)?;
            *val = resolved_schema;
        }
        match val {
            Value::Object(ref mut map) => {
                for v in map.values_mut() {
                    Self::resolve_references(schemas, v)?;
                }
            }
            Value::Array(ref mut arr) => {
                for v in arr.iter_mut() {
                    Self::resolve_references(schemas, v)?;
                }
            }
            _ => (),
        }
        Ok(())
    }

    /// Extract the fragment and get the target field name
    #[inline]
    fn extract_ref(val: &Value) -> Option<String> {
        let refr = val.get(Self::REF)?;
        let Value::String(final_refr) = refr else {
            return None;
        };
        final_refr.split('/').last().map(ToOwned::to_owned)
    }

    /// Inspect (validate) the given JSON value to check if it conforms to the compiled schema
    ///
    /// [`SchemaError`] is returned upon failure which contains a detailed explanation of everything
    /// that is inconsistent with the compiled schema.
    pub fn inspect(&self, json: &Value) -> Result<(), SchemaError> {
        self.schema.validate(json).map_err(|err| {
            let context = err
                .map(|err| ErrContext {
                    field: err.instance_path.to_string(),
                    description: err.to_string(),
                })
                .collect();

            SchemaError { context }
        })
    }
}
